/*
 * Copyright (c) 1998, 1999, 2000, 2001
 *	Phil Thompson <phil@river-bank.demon.co.uk>
 * Copyright (c) 2002
 *	Riverbank Computing Limited <info@riverbankcomputing.co.uk>
 *
 * The SIP parser.
 */

%{
#include <stdlib.h>
#include <string.h>

#include "sip.h"


#define	MAX_NESTED_IF		10
#define	MAX_NESTED_SCOPE	10

#define	isMainModule(m)		(currentSpec -> module == (m))


/* The different attribute name types. */

typedef enum {
	func_attr,
	var_attr,
	enum_attr,
	class_attr
} attrType;


static sipSpec *currentSpec;		/* The current spec being parsed. */
static stringList *neededVersions;	/* The list of required versions. */
static moduleDef *currentModule;	/* The current module being parsed. */
static mappedTypeDef *currentMappedType;	/* The current mapped type. */
static enumDef *currentEnum;		/* The current enum being parsed. */
static int sectionFlags;		/* The current section flags. */
static int currentOverIsVirt;		/* Set if the overload is virtual. */
static int currentIsStatic;		/* Set if the current is static. */
static char *previousFile;		/* The file just parsed. */
static parserContext newContext;	/* The new pending context. */
static int currentVersScope;		/* The index of the version scope. */
static versionRange versScope[MAX_NESTED_IF];	/* Current version nest. */
static classDef *scopeStack[MAX_NESTED_SCOPE];	/* The scope stack. */
static int sectFlagsStack[MAX_NESTED_SCOPE];	/* The section flags stack. */
static int currentScopeIdx;		/* The scope stack index. */
static versionDef currentFeature;	/* The current feature. */
static versionDef *currentVersion;	/* The current version. */


static nameDef *cacheName(sipSpec *,moduleDef *,char *);
static ifaceFileDef *findIfaceFile(sipSpec *,scopedNameDef *,ifaceFileType,
				   argDef *);
static classDef *findClass(sipSpec *,ifaceFileType,scopedNameDef *);
static classDef *newClass(sipSpec *,ifaceFileType,scopedNameDef *,versionDef *);
static void finishClass(classDef *);
static mappedTypeDef *newMappedType(sipSpec *,argDef *,versionDef *);
static void setCurrentVersion(void);
static enumDef *newEnum(sipSpec *,moduleDef *,char *,int);
static void newTypedef(sipSpec *,char *,argDef *);
static void newVar(sipSpec *,versionDef *,moduleDef *,char *,int,argDef *,
		   codeBlock *);
static void newCtor(versionDef *,char *,int,funcArgs *,optFlags *,codeBlock *);
static void newFunction(sipSpec *,versionDef *,moduleDef *,int,int,int,
			argDef *,char *,funcArgs *,int,int,optFlags *,
			codeBlock *,codeBlock *);
static optFlag *findOptFlag(optFlags *,char *,flagType);
static memberDef *findFunction(sipSpec *,moduleDef *,classDef *,char *);
static void checkAttributes(sipSpec *,classDef *,nameDef *,attrType);
static void newModule(FILE *,char *);
static void appendCodeBlock(codeBlock **,codeBlock *);
static void parseFile(FILE *,char *,moduleDef *);
static void handleEOF(void);
static void handleEOM(void);
static versionQual *findQualifier(sipSpec *,char *);
static versionQual *findPlatOrFeature(sipSpec *,char *);
static versionQual *findFeature(sipSpec *,char *);
static scopedNameDef *text2scopedName(char *);
static void pushScope(classDef *);
static void popScope(void);
static classDef *currentScope(void);
static void newVersion(sipSpec *,moduleDef *,int,char *,char *,versionType);
static void newImport(sipSpec *,char *);
static void usedInMainModule(sipSpec *,ifaceFileDef *);
static void orVersionQual(versionOrList **,versionQual *);
static void andVersionOrList(versionAndList **,versionOrList *);
static int isNeeded(versionQual *);
%}

%union {
	char		qchar;
	char		*text;
	long		number;
	double		real;
	argDef		memArg;
	funcArgs	arglist;
	codeBlock	*codeb;
	valueDef	value;
	valueDef	*valp;
	optFlags	optflags;
	optFlag		flag;
	scopedNameDef	*scpvalp;
	fcallDef	fcall;
	versionRange	versrange;
	versionOrList	*oredlist;
}

%token			TK_DOC
%token			TK_EXPORTEDDOC
%token			TK_MAKEFILE
%token			TK_VARCODE
%token			TK_POSTINITCODE
%token			TK_CPPCODE
%token			TK_PREPYCODE
%token			TK_PYCODE
%token 			TK_COPYING
%token			TK_MAPPEDTYPE
%token <codeb>		TK_CODEBLOCK
%token			TK_IF
%token			TK_END
%token <text>		TK_NAME
%token <text>		TK_FILENAME
%token <text>		TK_STRING
%token			TK_VIRTUALCODE
%token			TK_MEMBERCODE
%token			TK_FROMTYPE
%token			TK_TOTYPE
%token			TK_TOSUBCLASS
%token			TK_INCLUDE
%token			TK_IMPORT
%token			TK_FEATURECODE
%token			TK_EXPFEATURESCODE
%token			TK_HEADERCODE
%token			TK_EXPHEADERCODE
%token			TK_MODULE
%token			TK_CLASS
%token			TK_STRUCT
%token			TK_PUBLIC
%token			TK_PROTECTED
%token			TK_PRIVATE
%token			TK_SIGNALS
%token			TK_SLOTS
%token			TK_BOOL
%token			TK_SHORT
%token			TK_INT
%token			TK_LONG
%token			TK_FLOAT
%token			TK_DOUBLE
%token			TK_CHAR
%token			TK_VOID
%token			TK_VIRTUAL
%token			TK_ENUM
%token			TK_UNSIGNED
%token			TK_SCOPE
%token			TK_LOGICAL_OR
%token			TK_CONST
%token			TK_STATIC
%token			TK_SIPSIGNAL
%token			TK_SIPRXCON
%token			TK_SIPRXDIS
%token			TK_SIPSLOTCON
%token			TK_SIPSLOTDIS
%token <number>		TK_NUMBER
%token <real>		TK_REAL
%token			TK_TYPEDEF
%token			TK_NAMESPACE
%token			TK_TIMELINE
%token			TK_PLATFORMS
%token			TK_FEATURE
%token			TK_EXPOSE
%token <qchar>		TK_QCHAR

%type <memArg>		argvalue
%type <memArg>		argtype
%type <memArg>		cpptype
%type <memArg>		basetype
%type <arglist>		arglist
%type <arglist>		rawarglist
%type <arglist>		basetypelist
%type <number>		optslot
%type <number>		optref
%type <number>		optconst
%type <number>		optabstract
%type <number>		deref
%type <value>		simplevalue
%type <valp>		value
%type <valp>		expr
%type <valp>		optassign
%type <codeb>		optvarcode
%type <codeb>		hdrcode
%type <codeb>		cppcode
%type <codeb>		codeblock
%type <codeb>		optvirtualcode
%type <codeb>		membercode
%type <text>		optfilename
%type <text>		optname
%type <optflags>	optflags
%type <optflags>	flaglist
%type <flag>		flag
%type <flag>		flagvalue
%type <qchar>		optunop
%type <qchar>		binop
%type <scpvalp>		scopepart
%type <scpvalp>		scopedname
%type <fcall>		valuelist
%type <versrange>	versionrange
%type <oredlist>	oredversions

%%

specification: 	statement
	|	specification statement
	;

statement:	{
			/*
			 * We don't do these in parserEOF() because the parser
			 * is reading ahead and that would be too early.
			 */

			if (previousFile != NULL)
			{
				handleEOF();

				if (newContext.prevmod != NULL)
					handleEOM();

				free(previousFile);
				previousFile = NULL;
			}
	} modstatement
	;

modstatement:	module
	|	copying
	|	include
	|	import
	|	timeline
	|	platforms
	|	feature
	|	expose
	|	featurecode
	|	expfeatcode
	|	hdrcode {
			if (isMainModule(currentModule))
				appendCodeBlock(&currentSpec -> hdrcode,$1);
		}
	|	exphdrcode
	|	cppcode {
			if (isMainModule(currentModule))
				appendCodeBlock(&currentSpec -> cppcode,$1);
		}
	|	postinitcode
	|	prepycode
	|	pycode
	|	doc
	|	exporteddoc
	|	makefile
	|	mappedtype
	|	nsstatement
	;

nsstatement:	ifstart
	|	ifend
	|	namespace
	|	class
	|	opaqueclass
	|	typedef
	|	enum
	|	function
	|	variable
	;

mappedtype:	TK_MAPPEDTYPE basetype {
			if (currentVersion != NULL)
			{
				$2.nrderefs = 0;
				$2.argflags = 0;

				currentMappedType = newMappedType(currentSpec,&$2,currentVersion);
				usedInMainModule(currentSpec,currentMappedType -> iff);
			}
		} '{' mtbody '}' ';' {
			if (currentVersion != NULL)
			{
				if (currentMappedType -> convfromcode == NULL)
					yyerror("%MappedType must have a %ConvertFromTypeCode directive");

				if (currentMappedType -> convtocode == NULL)
					yyerror("%MappedType must have a %ConvertToTypeCode directive");

				currentMappedType = NULL;
			}
		}
	;

mtbody:		mtline
	|	mtbody mtline
	;

mtline: 	hdrcode {
			if (currentVersion != NULL)
				appendCodeBlock(&currentMappedType -> hdrcode,$1);
		}
	|	TK_FROMTYPE codeblock {
			if (currentVersion != NULL)
			{
				if (currentMappedType -> convfromcode != NULL)
					yyerror("%MappedType has more than one %ConvertFromTypeCode directive");

				currentMappedType -> convfromcode = $2;
			}
		}
	|	TK_TOTYPE codeblock {
			if (currentVersion != NULL)
			{
				if (currentMappedType -> convtocode != NULL)
					yyerror("%MappedType has more than one %ConvertToTypeCode directive");

				currentMappedType -> convtocode = $2;
			}
		}
	;

namespace:	TK_NAMESPACE TK_NAME {
			if (currentVersion != NULL)
			{
				classDef *ns;

				ns = newClass(currentSpec,namespace_iface,text2scopedName($2),currentVersion);
			
				setCannotCreate(ns);

				pushScope(ns);

				sectionFlags = 0;
			}
		} '{' nsbody '}' ';' {
			/* Remove the namespace name. */

			if (currentVersion != NULL)
				popScope();
		}
	;

nsbody:		nsstatement
	|	nsbody nsstatement
	;

platforms:	TK_PLATFORMS {
			versionQual *vq;

			for (vq = currentModule -> versions; vq != NULL; vq = vq -> next)
				if (vq -> vtype == platform_version)
					yyerror("%Platforms has already been defined for this module");
		}
		'{' platformlist '}'
	;

platformlist:	platform
	|	platformlist platform
	;

platform:	TK_NAME {
			newVersion(currentSpec,currentModule,-1,$1,NULL,platform_version);
		}
	;

feature:	TK_FEATURE TK_NAME TK_STRING {
			newVersion(currentSpec,currentModule,-1,$2,$3,feature_version);
		}
	;

timeline:	TK_TIMELINE {
			if (currentModule -> nrTimeSlots >= 0)
				yyerror("%Timeline has already been defined for this module");

			currentModule -> nrTimeSlots = 0;
		}
		'{' versionlist '}'
	;

versionlist:	versionname
	|	versionlist versionname
	;

versionname:	TK_NAME {
			newVersion(currentSpec,currentModule,currentModule -> nrTimeSlots++,$1,NULL,time_version);
		}
	;

ifstart:	TK_IF '(' versionrange ')' {
			if (currentVersScope >= MAX_NESTED_IF)
				yyerror("Internal error: increase the value of MAX_NESTED_IF");

			versScope[currentVersScope++] = $3;

			setCurrentVersion();
		}
	;

oredversions:	TK_NAME {
			$$ = NULL;

			orVersionQual(&$$,findPlatOrFeature(currentSpec,$1));
		}
	|	oredversions TK_LOGICAL_OR TK_NAME {
			$$ = $1;

			orVersionQual(&$$,findPlatOrFeature(currentSpec,$3));
		}
	;

versionrange:	oredversions {
			$$.lower = $$.upper = NULL;
			$$.ored = $1;
		}
	|	optname '-' optname {
			versionQual *vq;

			$$.ored = NULL;

			if ($1 == NULL)
				$$.lower = NULL;
			else if ((vq = findQualifier(currentSpec,$1)) == NULL || vq -> vtype != time_version)
				yyerror("Lower bound is not a time version");
			else
				$$.lower = vq;

			if ($3 == NULL)
				$$.upper = NULL;
			else if ((vq = findQualifier(currentSpec,$3)) == NULL || vq -> vtype != time_version)
				yyerror("Upper bound is not a time version");
			else
				$$.upper = vq;

			/* Sanity checks on the bounds. */

			if ($$.lower != NULL && $$.upper != NULL)
			{
				if ($$.lower -> module != $$.upper -> module)
					yyerror("Lower and upper bounds are from different timelines");

				if ($$.lower == $$.upper)
					yyerror("Lower and upper bounds must be different");

				if ($$.lower -> order > $$.upper -> order)
					yyerror("Later version specified as lower bound");
			}
		}
	;

ifend:		TK_END {
			if (currentVersScope-- <= 0)
				yyerror("Too many %End directives");

			setCurrentVersion();
		}
	;

expose:		TK_EXPOSE TK_NAME {
			/* Only bother if this is the main module. */

			if (isMainModule(currentModule))
			{
				expFuncDef *ef;

				/* Check it isn't already there. */

				for (ef = currentSpec -> exposed; ef != NULL; ef = ef -> next)
					if (strcmp(ef -> name,$2) == 0)
						yyerror("Function has already been exposed in this module");

				ef = sipMalloc(sizeof (expFuncDef));

				ef -> name = $2;
				ef -> next = currentSpec -> exposed;

				currentSpec -> exposed = ef;
			}
		}
	;

module:		TK_MODULE TK_NAME {
			/* Check the module hasn't already been defined. */

			moduleDef *mod;

			for (mod = currentSpec -> moduleList; mod != NULL; mod = mod -> next)
				if (mod -> name != NULL && strcmp(mod -> name,$2) == 0)
					yyerror("Module is already defined");

			currentModule -> name = $2;
		}
	;

include:	TK_INCLUDE TK_FILENAME {
			parseFile(NULL,$2,NULL);
		}
	;

import:		TK_IMPORT TK_FILENAME {
			newImport(currentSpec,$2);
		}
	;

optvarcode:	{
			$$ = NULL;
		}
	|	TK_VARCODE codeblock {
			$$ = $2;
		}
	;

featurecode:	TK_FEATURECODE codeblock {
			appendCodeBlock(&currentSpec -> featurecode,$2);
		}
	;

expfeatcode:	TK_EXPFEATURESCODE codeblock {
			appendCodeBlock(&currentSpec -> expfeatcode,$2);
		}
	;

copying:	TK_COPYING codeblock {
			if (isMainModule(currentModule))
				appendCodeBlock(&currentSpec -> copying,$2);
		}
	;

hdrcode:	TK_HEADERCODE codeblock {
			$$ = $2;
		}
	;

exphdrcode:	TK_EXPHEADERCODE codeblock {
			appendCodeBlock(&currentSpec -> exphdrcode,$2);
		}
	;

cppcode:	TK_CPPCODE codeblock {
			$$ = $2;
		}
	;

postinitcode:	TK_POSTINITCODE codeblock {
			if (isMainModule(currentModule))
				appendCodeBlock(&currentSpec -> postinitcode,$2);
		}
	;

prepycode:	TK_PREPYCODE codeblock {
			if (isMainModule(currentModule))
				appendCodeBlock(&currentSpec -> prepycode,$2);
		}
	;

pycode:		TK_PYCODE codeblock {
			if (isMainModule(currentModule))
				appendCodeBlock(&currentSpec -> pycode,$2);
		}
	;

doc:		TK_DOC codeblock {
			if (isMainModule(currentModule))
				appendCodeBlock(&currentSpec -> docs,$2);
		}
	;

exporteddoc:	TK_EXPORTEDDOC codeblock {
			appendCodeBlock(&currentSpec -> docs,$2);
		}
	;

makefile:	TK_MAKEFILE TK_FILENAME optfilename codeblock {
			if (isMainModule(currentModule))
			{
				mkTemplateDef *mtd;

				mtd = sipMalloc(sizeof (mkTemplateDef));

				mtd -> name = $2;

				if ((mtd -> objext = $3) == NULL)
					mtd -> objext = DEFAULT_OFILE_EXT;

				mtd -> templ = NULL;

				appendCodeBlock(&mtd -> templ,$4);

				mtd -> next = currentSpec -> mktemplates;
				currentSpec -> mktemplates = mtd;
			}
		}
	;

codeblock:	TK_CODEBLOCK TK_END
	;

enum:		TK_ENUM optname {
			if (currentVersion != NULL)
			{
				if (sectionFlags != 0 && (sectionFlags & ~(SECT_IS_PUBLIC | SECT_IS_PROT)) != 0)
					yyerror("Class enums must be in the public or protected sections");

				currentEnum = newEnum(currentSpec,currentModule,$2,sectionFlags);
			}
		} '{' enumbody '}' ';'
	;

optfilename:	{
			$$ = NULL;
		}
	|	TK_FILENAME {
			$$ = $1;
		}
	;

optname:	{
			$$ = NULL;
		}
	|	TK_NAME {
			$$ = $1;
		}
	;

enumbody:	enumline
	|	enumbody enumline
	;

enumline:	ifstart
	|	ifend
	|	TK_NAME optassign optcomma {
			if (currentVersion != NULL)
			{
				/*
				 * Note that we don't use the assigned value.
				 * This is a hangover from when enums where
				 * generated in Python.  We can remove it when
				 * we have got around to updating all the .sip
				 * files.
				 */

				enumValueDef *evd;

				evd = sipMalloc(sizeof (enumValueDef));

				evd -> name = cacheName(currentSpec,currentModule,$1);
				evd -> version = *currentVersion;
				evd -> ed = currentEnum;

				checkAttributes(currentSpec,evd -> ed -> ecd,evd -> name,enum_attr);

				evd -> next = currentEnum -> values;
				currentEnum -> values = evd;

			}
		}
	;

optcomma:
	|	','
	;

optassign:	{
			$$ = NULL;
		}
	|	'=' expr {
			$$ = $2;
		}
	;

expr:		value
	|	expr binop value {
			valueDef *vd;
 
			if ($1 -> vtype == string_value || $3 -> vtype == string_value)
				yyerror("Invalid binary operator for string");
 
			/* Find the last value in the existing expression. */
 
			for (vd = $1; vd -> next != NULL; vd = vd -> next)
				;
 
			vd -> vbinop = $2;
			vd -> next = $3;

			$$ = $1;
		}
	;

binop:		'-' {
			$$ = '-';
		}
	|	'+' {
			$$ = '+';
		}
	|	'|' {
			$$ = '|';
		}
	;

optunop:	{
			$$ = '\0';
		}
	|	'!' {
			$$ = '!';
		}
	|	'~' {
			$$ = '~';
		}
	|	'-' {
			$$ = '-';
		}
	|	'+' {
			$$ = '+';
		}
	;

value:		optunop simplevalue {
			if ($1 != '\0' && $2.vtype == string_value)
				yyerror("Invalid unary operator for string");
 
			/*
			 * Convert the value to a simple expression on the
			 * heap.
			 */
 
			$$ = sipMalloc(sizeof (valueDef));
 
			*$$ = $2;
			$$ -> vunop = $1;
			$$ -> vbinop = '\0';
			$$ -> next = NULL;
		}
	;

scopedname:	scopepart
	|	scopedname TK_SCOPE scopepart {
			appendScopedName(&$1,$3);
		}
	;

scopepart: 	TK_NAME {
			$$ = text2scopePart($1);
		}
	;

simplevalue:	scopedname {
			/*
			 * We let the C++ compiler decide if the value is a
			 * valid one - no point in building a full C++ parser
			 * here.
			 */

			$$.vtype = scoped_value;
			$$.u.vscp = $1;
		}
	|	scopedname '(' valuelist ')' {
			fcallDef *fcd;

			fcd = sipMalloc(sizeof (fcallDef));
			*fcd = $3;
			fcd -> name = $1;

			$$.vtype = fcall_value;
			$$.u.fcd = fcd;
		}
	|	TK_REAL {
			$$.vtype = real_value;
			$$.u.vreal = $1;
		}
	|	TK_NUMBER {
			$$.vtype = numeric_value;
			$$.u.vnum = $1;
		}
	|	TK_STRING {
			$$.vtype = string_value;
			$$.u.vstr = $1;
		}
	|	TK_QCHAR {
			$$.vtype = qchar_value;
			$$.u.vqchar = $1;
		}
	;

valuelist:	{
			/* No values. */

			$$.nrArgs = 0;
		}
	|	value {
			/* The single or first value. */

			$$.args[0] = $1;
			$$.nrArgs = 1;
		}
	|	valuelist ',' value {
			/* Check that it wasn't ...(,value...). */

			if ($$.nrArgs == 0)
				yyerror("First argument to function call is missing");

			/* Check there is room. */

			if ($1.nrArgs == MAX_NR_ARGS)
				yyerror("Too many arguments to function call");

			$$ = $1;

			$$.args[$$.nrArgs] = $3;
			$$.nrArgs++;
		}
	;

typedef:	TK_TYPEDEF cpptype TK_NAME ';' {
			if (currentVersion != NULL)
				newTypedef(currentSpec,$3,&$2);
		}
	;

opaqueclass:	TK_CLASS scopedname ';' {
			if (currentVersion != NULL)
			{
				classDef *cd;

				cd = newClass(currentSpec,class_iface,$2,currentVersion);

				usedInMainModule(currentSpec,cd -> iff);
				setCannotCreate(cd);
				setIsOpaque(cd);
			}
		}
	;

class:		TK_CLASS TK_NAME {
			if (currentVersion != NULL)
			{
				classDef *cd;

				cd = newClass(currentSpec,class_iface,text2scopedName($2),currentVersion);

				usedInMainModule(currentSpec,cd -> iff);
				setCannotCreate(cd);

				pushScope(cd);

				sectionFlags = SECT_IS_PRIVATE;
			}
		} superclasses '{' classbody '}' ';' {
			if (currentVersion != NULL)
			{
				finishClass(currentScope());
				popScope();
			}
		}
	;

superclasses:
	|	':' superlist
	;

superlist:	superclass
	|	superlist ',' superclass
	;

superclass:	scopedname {
			if (currentVersion != NULL)
			{
				classDef *cd, *super;

				cd = currentScope();

				super = findClass(currentSpec,class_iface,$1);

				appendToClassList(&cd -> supers,super);
				addToUsedList(&cd -> iff -> used,super -> iff);
			}
		}
	;

classbody:	classline
	|	classbody classline
	;

classline:	ifstart
	|	ifend
	|	namespace
	|	class
	|	opaqueclass
	|	typedef
	|	enum
	|	cppcode {
			if (currentVersion != NULL)
				appendCodeBlock(&currentScope() -> cppcode,$1);
		}
	|	hdrcode {
			if (currentVersion != NULL)
				appendCodeBlock(&currentScope() -> hdrcode,$1);
		}
	|	ctor
	|	dtor
	|	varmember
	|	TK_TOSUBCLASS codeblock {
			if (currentVersion != NULL)
			{
				classDef *cvd = currentScope();

				if (cvd -> convtosubcode != NULL)
					yyerror("Class has more than one %ConvertToSubClassCode directive");

				cvd -> convtosubcode = $2;
			}
		}
	|	TK_TOTYPE codeblock {
			if (currentVersion != NULL)
			{
				classDef *cvd = currentScope();

				if (cvd -> convtocode != NULL)
					yyerror("Class has more than one %ConvertToTypeCode directive");

				cvd -> convtocode = $2;
			}
		}
	|	TK_PUBLIC optslot ':' {
			if (currentVersion != NULL)
				sectionFlags = SECT_IS_PUBLIC | $2;
		}
	|	TK_PROTECTED optslot ':' {
			if (currentVersion != NULL)
				sectionFlags = SECT_IS_PROT | $2;
		}
	|	TK_PRIVATE optslot ':' {
			if (currentVersion != NULL)
				sectionFlags = SECT_IS_PRIVATE | $2;
		}
	|	TK_SIGNALS ':' {
			if (currentVersion != NULL)
				sectionFlags = SECT_IS_SIGNAL;
		}
	;

optslot:	{
			$$ = 0;
		}
	|	TK_SLOTS {
			$$ = SECT_IS_SLOT;
		}
	;

dtor:		'~' TK_NAME '(' ')' ';' membercode {
			if (currentVersion != NULL)
			{
				classDef *cvd = currentScope();

				if (strcmp(classBaseName(cvd),$2) != 0)
					yyerror("Destructor doesn't have the same name as its class");

				if (isDtor(cvd))
					yyerror("Destructor has already been defined");

				cvd -> dtorcode = $6;
				setIsDtor(cvd);

				if ((sectionFlags & SECT_IS_PUBLIC) == 0)
					setNoPublicDtor(cvd);
			}
		}
	;

ctor:		TK_NAME '(' arglist ')' optflags ';' membercode {
			if (currentVersion != NULL)
			{
				if ((sectionFlags & (SECT_IS_PUBLIC | SECT_IS_PROT | SECT_IS_PRIVATE)) == 0)
					yyerror("Constructor must be in the public, private or protected sections");

				newCtor(currentVersion,$1,sectionFlags,&$3,&$5,$7);
			}

			free($1);
		}
	;

function:	cpptype TK_NAME '(' arglist ')' optconst optabstract optflags ';' membercode optvirtualcode {
			if (currentVersion != NULL)
			{
				if (sectionFlags != 0 && (sectionFlags & (SECT_IS_PUBLIC | SECT_IS_PROT | SECT_IS_PRIVATE | SECT_IS_SLOT | SECT_IS_SIGNAL)) == 0)
					yyerror("Class function must be in the public, private, protected, slot or signal sections");

				newFunction(currentSpec,currentVersion,
					    currentModule,sectionFlags,
					    currentIsStatic,currentOverIsVirt,
					    &$1,$2,&$4,$6,$7,&$8,$10,$11);
			}

			currentIsStatic = FALSE;
			currentOverIsVirt = FALSE;
		}
	;


optconst:	{
			$$ = FALSE;
		}
	|	TK_CONST {
			$$ = TRUE;
		}
	;

optabstract:	{
			$$ = 0;
		}
	|	'=' TK_NUMBER {
			if ($2 != 0)
				yyerror("Abstract virtual function '= 0' expected");

			$$ = TRUE;
		}
	;

optflags:	{
			$$.nrFlags = 0;
		}
	|	'/' flaglist '/' {
			$$ = $2;
		}
	;


flaglist:	flag {
			$$.flags[0] = $1;
			$$.nrFlags = 1;
		}
	|	flaglist ',' flag {
			/* Check there is room. */

			if ($1.nrFlags == MAX_NR_FLAGS)
				yyerror("Too many optional flags");

			$$ = $1;

			$$.flags[$$.nrFlags++] = $3;
		}
	;

flag:		TK_NAME {
			$$.ftype = bool_flag;
			$$.fname = $1;
		}
	|	TK_NAME '=' flagvalue {
			$$ = $3;
			$$.fname = $1;
		}
	;

flagvalue:	TK_NAME {
			$$.ftype = name_flag;
			$$.fvalue.sval = $1;
		}
	|	TK_STRING {
			$$.ftype = string_flag;
			$$.fvalue.sval = $1;
		}
	;

membercode:	{
			$$ = NULL;
		}
	|	TK_MEMBERCODE codeblock {
			$$ = $2;
		}
	;

optvirtualcode:	{
			$$ = NULL;
		}
	|	TK_VIRTUALCODE codeblock {
			$$ = $2;
		}
	;

arglist:	rawarglist {
			int a, nrsigs, nrrxcon, nrrxdis, nrslotcon, nrslotdis, nrarray, nrarraysize;

			nrsigs = nrrxcon = nrrxdis = nrslotcon = nrslotdis = nrarray = nrarraysize = 0;

			for (a = 0; a < $1.nrArgs; ++a)
			{
				argDef *ad = &$1.args[a];

				switch (ad -> atype)
				{
				case signal_type:
					++nrsigs;
					break;

				case rxcon_type:
					++nrrxcon;
					break;

				case rxdis_type:
					++nrrxdis;
					break;

				case slotcon_type:
					++nrslotcon;
					break;

				case slotdis_type:
					++nrslotdis;
					break;
				}

				if (isArray(ad))
					++nrarray;

				if (isArraySize(ad))
					++nrarraysize;
			}

			if (nrsigs == 1 && nrrxcon == 0)
				yyerror("SIP_SIGNAL can only be used with SIP_RXOBJ_CON");

			if (nrsigs > 1)
				yyerror("SIP_SIGNAL can only be given once");

			if (nrrxcon != nrslotcon || nrrxcon > 1)
				yyerror("SIP_RXOBJ_CON and SIP_SLOT_CON must both be given and at most once");

			if (nrrxdis != nrslotdis || nrrxdis > 1)
				yyerror("SIP_RXOBJ_DIS and SIP_SLOT_DIS must both be given and at most once");

			if (nrarray != nrarraysize || nrarray > 1)
				yyerror("/Array/ and /ArraySize/ must both be given and at most once");

			$$ = $1;
		}
	;

rawarglist:	{
			/* No arguments. */

			$$.nrArgs = 0;
		}
	|	argvalue {
			/* The single or first argument. */

			$$.args[0] = $1;
			$$.nrArgs = 1;
		}
	|	rawarglist ',' argvalue {
			/* Check that it wasn't ...(,arg...). */

			if ($$.nrArgs == 0)
				yyerror("First argument of list is missing");

			/*
			 * If this argument has no default value, then all
			 * previous ones mustn't either.
			 */

			if ($3.defval == NULL)
			{
				int a;

				for (a = 0; a < $1.nrArgs; ++a)
					if ($1.args[a].defval != NULL)
						yyerror("Compulsory argument given after optional argument");
			}

			/* Check there is room. */

			if ($1.nrArgs == MAX_NR_ARGS)
				yyerror("Internal error - increase the value of MAX_NR_ARGS");

			$$ = $1;

			$$.args[$$.nrArgs] = $3;
			$$.nrArgs++;
		}
	;

argvalue:	TK_SIPSIGNAL {
			$$.atype = signal_type;
			$$.argflags = 0;
			$$.nrderefs = 0;

			currentSpec -> sigslots = TRUE;
		}
	|	TK_SIPRXCON {
			$$.atype = rxcon_type;
			$$.argflags = 0;
			$$.nrderefs = 0;

			currentSpec -> sigslots = TRUE;
		}
	|	TK_SIPRXDIS {
			$$.atype = rxdis_type;
			$$.argflags = 0;
			$$.nrderefs = 0;

			currentSpec -> sigslots = TRUE;
		}
	|	TK_SIPSLOTCON '(' arglist ')' {
			$$.atype = slotcon_type;
			$$.argflags = 0;
			$$.nrderefs = 0;

			$$.u.sa = sipMalloc(sizeof (funcArgs));
			*$$.u.sa = $3;

			currentSpec -> sigslots = TRUE;
		}
	|	TK_SIPSLOTDIS '(' arglist ')' {
			$$.atype = slotdis_type;
			$$.argflags = 0;
			$$.nrderefs = 0;

			$$.u.sa = sipMalloc(sizeof (funcArgs));
			*$$.u.sa = $3;

			currentSpec -> sigslots = TRUE;
		}
	|	argtype optassign {
			$$ = $1;
			$$.defval = $2;
		}
	;

varmember:	TK_STATIC {currentIsStatic = TRUE;} varmem
	|	varmem
	;

varmem:		member
	|	variable
	;

member:		TK_VIRTUAL {currentOverIsVirt = TRUE;} function
	|	function
	;

variable:	cpptype TK_NAME ';' optvarcode {
			if (currentVersion != NULL)
			{
				/* Check the section. */

				if (sectionFlags != 0)
				{
					if ((sectionFlags & SECT_IS_PUBLIC) == 0)
						yyerror("Class variables must be in the public section");

					if (!currentIsStatic && $4 != NULL)
						yyerror("%VariableCode cannot be specified for non-static class variables");
				}

				newVar(currentSpec,currentVersion,
				       currentModule,$2,currentIsStatic,&$1,$4);
			}

			currentIsStatic = FALSE;
		}
	;

cpptype:	TK_CONST basetype deref optref {
			$$ = $2;
			$$.nrderefs = $3;
			$$.argflags = ARG_IS_CONST | $4;
		}
	|	basetype deref optref {
			$$ = $1;
			$$.nrderefs = $2;
			$$.argflags = $3;
		}
	;

argtype:	cpptype optflags {
			$$ = $1;

			if (findOptFlag(&$2,"Array",bool_flag) != NULL)
				$$.argflags |= ARG_ARRAY;

			if (findOptFlag(&$2,"ArraySize",bool_flag) != NULL)
				$$.argflags |= ARG_ARRAY_SIZE;

			if (findOptFlag(&$2,"Transfer",bool_flag) != NULL)
				$$.argflags |= ARG_XFERRED;

			if (findOptFlag(&$2,"TransferThis",bool_flag) != NULL)
				$$.argflags |= ARG_THIS_XFERRED;

			if (findOptFlag(&$2,"TransferBack",bool_flag) != NULL)
				$$.argflags |= ARG_XFERRED_BACK;

			if (findOptFlag(&$2,"Constrained",bool_flag) != NULL)
			{
				if ($$.atype != int_type)
					yyerror("/Constrained/ may only be specified for int types");

				$$.atype = cint_type;
			}
		}
	;

optref:		{
			$$ = 0;
		}
	|	'&' {
			$$ = ARG_IS_REF;
		}
	;

deref:		{
			$$ = 0;
		}
	|	deref '*' {
			$$ = $1 + 1;
		}
	;

basetype:	scopedname {
			$$.atype = defined_type;
			$$.u.snd = $1;
		}
	|	scopedname '<' basetypelist '>' {
			templateDef *td;

			td = sipMalloc(sizeof(templateDef));
			td -> fqname = $1;
			td -> types = $3;

			$$.atype = template_type;
			$$.u.td = td;
		}
	|	TK_STRUCT scopedname {
			$$.atype = struct_type;
			$$.u.sname = $2;
		}
	|	TK_UNSIGNED TK_SHORT {
			$$.atype = ushort_type;
		}
	|	TK_SHORT {
			$$.atype = short_type;
		}
	|	TK_UNSIGNED {
			$$.atype = uint_type;
		}
	|	TK_INT {
			$$.atype = int_type;
		}
	|	TK_LONG {
			$$.atype = long_type;
		}
	|	TK_UNSIGNED TK_LONG {
			$$.atype = ulong_type;
		}
	|	TK_FLOAT {
			$$.atype = float_type;
		}
	|	TK_DOUBLE {
			$$.atype = double_type;
		}
	|	TK_BOOL {
			$$.atype = bool_type;
		}
	|	TK_UNSIGNED TK_CHAR {
			$$.atype = ustring_type;
		}
	|	TK_CHAR {
			$$.atype = string_type;
		}
	|	TK_VOID {
			$$.atype = voidptr_type;
		}
	;

basetypelist:	basetype {
			/* The single or first type. */

			$$.args[0] = $1;
			$$.nrArgs = 1;
		}
	|	basetypelist ',' basetype {
			/* Check that it wasn't ,type... */

			if ($$.nrArgs == 0)
				yyerror("First type of list is missing");

			/* Check there is room. */

			if ($1.nrArgs == MAX_NR_ARGS)
				yyerror("Internal error - increase the value of MAX_NR_ARGS");

			$$ = $1;

			$$.args[$$.nrArgs] = $3;
			$$.nrArgs++;
		}
	;

%%


/*
 * Parse the specification.
 */

void parse(sipSpec *spec,FILE *fp,char *filename,char *cppMName,stringList *tsl)
{
        /* Initialise the spec. */
 
	spec -> cppmname = cppMName;
	spec -> namecache = NULL;
	spec -> moduleList = NULL;
	spec -> ifacefiles = NULL;
	spec -> classes = NULL;
	spec -> mappedtypes = NULL;
	spec -> used = NULL;
	spec -> qobjclass = -1;
	spec -> exposed = NULL;
	spec -> enums = NULL;
	spec -> vars = NULL;
	spec -> othfuncs = NULL;
	spec -> overs = NULL;
	spec -> typedefs = NULL;
	spec -> featurecode = NULL;
	spec -> copying = NULL;
	spec -> hdrcode = NULL;
	spec -> expfeatcode = NULL;
	spec -> exphdrcode = NULL;
	spec -> cppcode = NULL;
	spec -> prepycode = NULL;
	spec -> pycode = NULL;
	spec -> docs = NULL;
	spec -> postinitcode = NULL;
	spec -> mktemplates = NULL;
	spec -> sigargs = NULL;
	spec -> sigslots = FALSE;

	currentSpec = spec;
	neededVersions = tsl;
	currentModule = NULL;
	currentMappedType = NULL;
	currentOverIsVirt = FALSE;
	currentIsStatic = FALSE;
	previousFile = NULL;
	currentVersScope = 0;
	currentScopeIdx = 0;
	sectionFlags = 0;

	setCurrentVersion();

	newModule(fp,filename);
	spec -> module = currentModule;

	yyparse();

	handleEOF();
	handleEOM();
}


/*
 * Tell the parser that a complete file has now been read.
 */

void parserEOF(char *name,parserContext *pc)
{
	previousFile = sipStrdup(name);
	newContext = *pc;
}


/*
 * Append a class definition to a class list.  Append is needed specifically
 * for the list of super-classes because the order is important to Python.
 */

void appendToClassList(classList **clp,classDef *cd)
{
	classList *new;

	/* Find the end of the list. */

	while (*clp != NULL)
		clp = &(*clp) -> next;

	new = sipMalloc(sizeof (classList));

	new -> cd = cd;
	new -> next = NULL;

	*clp = new;
}


/*
 * Create a new module for the current specification and make it current.
 */

static void newModule(FILE *fp,char *filename)
{
	parseFile(fp,filename,currentModule);

	currentModule = sipMalloc(sizeof (moduleDef));

	currentModule -> name = NULL;
	currentModule -> file = filename;
	currentModule -> versions = NULL;
	currentModule -> nrTimeSlots = -1;
	currentModule -> next = currentSpec -> moduleList;

	currentSpec -> moduleList = currentModule;
}


/*
 * Switch to parsing a new file.
 */

static void parseFile(FILE *fp,char *name,moduleDef *prevmod)
{
	parserContext pc;

	pc.ifdepth = currentVersScope;
	pc.prevmod = prevmod;

	setInputFile(fp,name,&pc);
}


/*
 * Find an interface file, or create a new one.
 */

static ifaceFileDef *findIfaceFile(sipSpec *pt,scopedNameDef *fqname,
				   ifaceFileType iftype,argDef *ad)
{
	ifaceFileDef *iff;

	/* See if the name is already used. */

	for (iff = pt -> ifacefiles; iff != NULL; iff = iff -> next)
	{
		if (!sameScopedName(iff -> fqname,fqname))
			continue;

		/* They must be the same type. */

		if (iff -> type != iftype)
			yyerror("A class, namespace or mapped type has already been defined with the same name");

		/*
		 * If this is a mapped type with the same name defined in a
		 * different module, then check that this type isn't the same
		 * as any of the mapped types defined in that module.
		 */

		if (iftype == mappedtype_iface && iff -> module != currentModule)
		{
			mappedTypeDef *mtd;

			for (mtd = pt -> mappedtypes; mtd != NULL; mtd = mtd -> next)
			{
				if (mtd -> iff != iff)
					continue;

				if (ad -> atype != template_type ||
				    mtd -> type.atype != template_type ||
				    sameBaseType(ad,&mtd -> type,TRUE))
					yyerror("Mapped type has already been defined in another module");
			}

			/*
			 * If we got here then we have a mapped type based on
			 * an existing template, but with unique parameters.
			 * We don't want to use interface files from other
			 * modules, so skip this one.
			 */

			continue;
		}

		return iff;
	}

	iff = sipMalloc(sizeof (ifaceFileDef));

	iff -> name = cacheName(pt,currentModule,scopedNameTail(fqname));
	iff -> type = iftype;
	iff -> fqname = fqname;
	iff -> module = NULL;
	iff -> ifacefileflags = 0;
	iff -> used = NULL;
	iff -> next = pt -> ifacefiles;

	pt -> ifacefiles = iff;

	return iff;
}


/*
 * Find a class definition in a parse tree.
 */

static classDef *findClass(sipSpec *pt,ifaceFileType iftype,
			   scopedNameDef *fqname)
{
	classDef *cd;
	ifaceFileDef *iff;

	/* See if the name is already used. */

	iff = findIfaceFile(pt,fqname,iftype,NULL);

	for (cd = pt -> classes; cd != NULL; cd = cd -> next)
		if (cd -> iff == iff)
			return cd;

	/* Create a new one. */

	cd = sipMalloc(sizeof (classDef));

	cd -> iff = iff;
	cd -> next = pt -> classes;

	pt -> classes = cd;

	return cd;
}


/*
 * Add an interface file to an interface file list if it isn't already there.
 */

void addToUsedList(ifaceFileList **ifflp,ifaceFileDef *iff)
{
	ifaceFileList *iffl;

	while ((iffl = *ifflp) != NULL)
	{
		/* Don't bother if it is already there. */

		if (iffl -> iff == iff)
			return;

		ifflp = &iffl -> next;
	}

	setIsUsed(iff);

	iffl = sipMalloc(sizeof (ifaceFileList));

	iffl -> iff = iff;
	iffl -> next = NULL;

	*ifflp = iffl;
}


/*
 * Find an undefined (or create a new) class definition in a parse tree.
 */

static classDef *newClass(sipSpec *pt,ifaceFileType iftype,
			  scopedNameDef *fqname,versionDef *vd)
{
	classDef *cd;

	cd = findClass(pt,iftype,fqname);

	/* Check it hasn't already been defined. */

	if (cd -> iff -> module != NULL)
		yyerror(iftype == class_iface ?
			"The class has already been defined" :
			"The namespace has already been defined");

	cd -> iff -> module = currentModule;

	/* Complete the initialisation. */

	cd -> classflags = 0;
	cd -> classnr = -1;
	cd -> pyindent = -1;
	cd -> version = *vd;
	cd -> ecd = currentScope();
	cd -> supers = NULL;
	cd -> hierachy = NULL;
	cd -> ctors = NULL;
	cd -> dtorcode = NULL;
	cd -> members = NULL;
	cd -> overs = NULL;
	cd -> vmembers = NULL;
	cd -> visible = NULL;
	cd -> cppcode = NULL;
	cd -> hdrcode = (cd -> ecd != NULL ? cd -> ecd -> hdrcode : NULL);
	cd -> convtosubcode = NULL;
	cd -> convtocode = NULL;

	return cd;
}


/*
 * Tidy up after finishing a class definition.
 */

static void finishClass(classDef *cvd)
{
	overDef *od;

	if (cvd -> ctors == NULL)
	{
		/* Provide a default ctor. */

		cvd -> ctors = sipMalloc(sizeof (ctorDef));
 
		cvd -> ctors -> ctorflags = SECT_IS_PUBLIC;
		cvd -> ctors -> args.nrArgs = 0;
		cvd -> ctors -> cppcode = NULL;
		cvd -> ctors -> version = cvd -> version;
		cvd -> ctors -> next = NULL;

		resetCannotCreate(cvd);
	}

	/* We can't create this class if it has private abstract functions. */

	for (od = cvd -> overs; od != NULL; od = od -> next)
		if (isAbstract(od) && isPrivate(od))
		{
			setCannotCreate(cvd);
			break;
		}
}


/*
 * Create a new mapped type.
 */

static mappedTypeDef *newMappedType(sipSpec *pt,argDef *ad,versionDef *vd)
{
	mappedTypeDef *mtd;
	scopedNameDef *snd;
	ifaceFileDef *iff;

	/* Check that the type is one we want to map. */

	switch (ad -> atype)
	{
	case defined_type:
		snd = ad -> u.snd;
		break;

	case template_type:
		snd = ad -> u.td -> fqname;
		break;

	case struct_type:
		snd = ad -> u.sname;
		break;

	default:
		yyerror("Invalid type for %MappedType");
	}

	iff = findIfaceFile(pt,snd,mappedtype_iface,ad);

	/* Check it hasn't already been defined. */

	for (mtd = pt -> mappedtypes; mtd != NULL; mtd = mtd -> next)
		if (mtd -> iff == iff)
		{
			/*
			 * We allow types based on the same template but with
			 * different arguments.
			 */

			if (ad -> atype != template_type ||
			    sameBaseType(ad,&mtd -> type,TRUE))
				yyerror("Mapped type has already been defined in this module");
		}

	/* The module may not have been set yet. */

	iff -> module = currentModule;

	/* Create a new mapped type. */

	mtd = sipMalloc(sizeof (mappedTypeDef));

	mtd -> type = *ad;
	mtd -> iff = iff;
	mtd -> version = *vd;
	mtd -> hdrcode = NULL;
	mtd -> convfromcode = NULL;
	mtd -> convtocode = NULL;
	mtd -> next = pt -> mappedtypes;

	pt -> mappedtypes = mtd;

	return mtd;
}


/*
 * Create a new enum.
 */

static enumDef *newEnum(sipSpec *pt,moduleDef *mod,char *name,int flags)
{
	enumDef *ed;
	classDef *escope = currentScope();

	ed = sipMalloc(sizeof (enumDef));

	ed -> enumflags = flags;
	ed -> fqname = (name != NULL ? text2scopedName(name) : NULL);
	ed -> ecd = escope;
	ed -> module = mod;
	ed -> values = NULL;
	ed -> next = pt -> enums;

	pt -> enums = ed;

	if (escope != NULL)
		setHasEnums(escope);

	return ed;
}


/*
 * Create a new typedef.
 */

static void newTypedef(sipSpec *pt,char *name,argDef *type)
{
	typedefDef *td;
	scopedNameDef *fqname = text2scopedName(name);

	/* Check it doesn't already exist. */

	for (td = pt -> typedefs; td != NULL; td = td -> next)
		if (sameScopedName(td -> fqname,fqname))
		{
			fatalScopedName(fqname);
			fatal(" already defined\n");
		}

	td = sipMalloc(sizeof (typedefDef));

	td -> fqname = fqname;
	td -> ecd = currentScope();
	td -> type = *type;
	td -> next = pt -> typedefs;

	pt -> typedefs = td;
}


/*
 * Create a new variable.
 */

static void newVar(sipSpec *pt,versionDef *vd,moduleDef *mod,char *name,
		   int isstatic,argDef *type,codeBlock *code)
{
	varDef *var;
	classDef *escope = currentScope();
	nameDef *nd = cacheName(pt,mod,name);

	checkAttributes(pt,escope,nd,var_attr);

	var = sipMalloc(sizeof (varDef));

	var -> name = nd;
	var -> fqname = text2scopedName(name);
	var -> ecd = escope;
	var -> module = mod;
	var -> varflags = 0;
	var -> type = *type;
	var -> version = *vd;
	var -> accessfunc = code;
	var -> next = pt -> vars;

	if (isstatic)
		setIsStaticVar(var);

	pt -> vars = var;
}


/*
 * Create a new ctor.
 */

static void newCtor(versionDef *vd,char *name,int sectFlags,funcArgs *args,
		    optFlags *optflgs,codeBlock *code)
{
	ctorDef *ct, **ctp;
	classDef *cd = currentScope();

	/* Check the name of the constructor. */

	if (strcmp(classBaseName(cd),name) != 0)
		yyerror("Constructor doesn't have the same name as its class");

	/*
	 * Find the end of the list and, at the same time, check for any
	 * existing one with the same signature and version.
	 */

	for (ctp = &cd -> ctors; *ctp != NULL; ctp = &(*ctp) -> next)
		if (sameFuncArgs(&(*ctp) -> args,args,FALSE))
			yyerror("A constructor with the same Python signature has already been defined");

	/* Add to the list of constructors. */

	ct = sipMalloc(sizeof (ctorDef));

	ct -> ctorflags = sectFlags;
	ct -> args = *args;
	ct -> cppcode = code;
	ct -> version = *vd;
	ct -> next = NULL;

	if (findOptFlag(optflgs,"ReleaseLock",bool_flag) != NULL)
		setIsBlockingCtor(ct);

	if (isProtectedCtor(ct))
		setIsComplex(cd);

	if (!isPrivateCtor(ct))
		resetCannotCreate(cd);

	*ctp = ct;
}


/*
 * Create a new function.
 */

static void newFunction(sipSpec *pt,versionDef *vd,moduleDef *mod,int sflags,
			int isstatic,int isvirt,argDef *type,char *name,
			funcArgs *args,int isconst,int isabstract,
			optFlags *optflgs,codeBlock *mcode,codeBlock *vcode)
{
	classDef *cvd = currentScope();
	char *pname;
	int a;
	overDef *od, **odp, **headp;
	optFlag *of;

	headp = (cvd != NULL ?  &cvd -> overs : &pt -> overs);

	/* Check /TransferThis/ wasn't specified. */

	for (a = 0; a < args -> nrArgs; ++a)
		if (isThisTransferred(&args -> args[a]))
			yyerror("/TransferThis/ may only be specified in constructors");

	/* Use the C++ name if a Python name wasn't given. */

	if ((of = findOptFlag(optflgs,"PyName",name_flag)) != NULL)
		pname = of -> fvalue.sval;
	else
		pname = name;

	/* Create a new overload definition. */

	od = sipMalloc(sizeof (overDef));

	/* Set the overload flags. */

	od -> overflags = sflags;

	if (isProtected(od))
		setIsComplex(cvd);

	if ((isSlot(od) || isSignal(od)) && !isPrivate(od))
	{
		if (isSignal(od))
			setIsComplex(cvd);

		pt -> sigslots = TRUE;
	}

	if (isSignal(od) && (mcode != NULL || vcode != NULL))
		yyerror("Cannot provide code for signals");

	if (isstatic)
	{
		if (isSlot(od) || isSignal(od))
			yyerror("Static functions cannot be slots or signals");

		if (isvirt)
			yyerror("Static functions cannot be virtual");

		setIsStatic(od);
	}

	if (isconst)
		setIsConst(od);

	if (isabstract)
	{
		if (sflags == 0)
			yyerror("Non-class function specified as abstract");

		if (mcode != NULL)
			yyerror("Cannot provide member code for abstract functions");

		setIsAbstract(od);
	}

	od -> agversion.val = NULL;

	if ((of = findOptFlag(optflgs,"AutoGen",opt_name_flag)) != NULL)
	{
		setIsAutoGen(od);

		if (of -> fvalue.sval != NULL)
		{
			versionOrList *newvol;
			versionAndList *val;

			/* Copy this one's version. */

			for (val = vd -> val; val != NULL; val = val -> next)
			{
				versionOrList *vol;

				newvol = NULL;

				for (vol = val -> vol; vol != NULL; vol = vol -> next)
					orVersionQual(&newvol,vol -> vq);

				if (newvol)
					andVersionOrList(&od -> agversion.val,newvol);
			}

			/* Add in the extra feature. */

			newvol = NULL;
			orVersionQual(&newvol,findFeature(pt,of -> fvalue.sval));
			andVersionOrList(&od -> agversion.val,newvol);
		}
	}

	if (findOptFlag(optflgs,"ReleaseLock",bool_flag) != NULL)
		setIsBlocking(od);

	if (isvirt)
	{
		if (isSignal(od))
			yyerror("Virtual signals aren't supported");

		setIsVirtual(od);
		setIsComplex(cvd);
	}
	else if (vcode != NULL)
		yyerror("Virtual code provided for non-virtual function");

	od -> cppname = name;
	od -> args = *args;
	od -> cppcode = mcode;
	od -> virtcode = vcode;
	od -> version = *vd;
	od -> common = findFunction(pt,mod,cvd,pname);

	/* Any overload being non-lazy makes all overloads non-lazy. */

	if (findOptFlag(optflgs,"NonLazy",bool_flag) != NULL)
		setIsNonLazy(od -> common);

	if ((of = findOptFlag(optflgs,"PreHook",name_flag)) != NULL)
		od -> prehook = of -> fvalue.sval;
	else
		od -> prehook = NULL;

	if ((of = findOptFlag(optflgs,"PostHook",name_flag)) != NULL)
		od -> posthook = of -> fvalue.sval;
	else
		od -> posthook = NULL;

	if (type -> atype == voidptr_type && type -> nrderefs == 0)
		od -> result = NULL;
	else
	{
		od -> result = sipMalloc(sizeof (argDef));
		*od -> result = *type;
	}

	od -> next = NULL;

	/*
	 * Find the end of the list and, at the same time, check for any
	 * existing one with the same signature and an overlapping version.
	 */

	for (odp = headp; *odp != NULL; odp = &(*odp) -> next)
	{
		if ((*odp) -> common != od -> common)
			continue;

		if (!sameOverload(*odp,od,FALSE))
			continue;

		yyerror("A function with the same Python signature has already been defined");
	}

	*odp = od;
}


/*
 * Cache a name in a module.
 */

static nameDef *cacheName(sipSpec *pt,moduleDef *mod,char *name)
{
	nameDef *nd;

	/* See if it already exists. */

	for (nd = pt -> namecache; nd != NULL; nd = nd -> next)
		if (strcmp(nd -> text,name) == 0)
			return nd;

	/* Create a new one. */

	nd = sipMalloc(sizeof (nameDef));

	nd -> text = name;
	nd -> module = mod;
	nd -> next = pt -> namecache;

	pt -> namecache = nd;

	return nd;
}


/*
 * Find (or create) an overloaded function name.
 */

static memberDef *findFunction(sipSpec *pt,moduleDef *mod,classDef *cvd,
			       char *pname)
{
	memberDef *md, **flist;
	nameDef *pnd;

	/* Cache the name. */

	pnd = cacheName(pt,mod,pname);

	/* Check there is no name clash. */

	checkAttributes(pt,cvd,pnd,func_attr);

	/* See if it already exists. */

	flist = (cvd != NULL ? &cvd -> members : &pt -> othfuncs);

	for (md = *flist; md != NULL; md = md -> next)
		if (md -> pyname == pnd)
			return md;

	/* Create a new one. */

	md = sipMalloc(sizeof (memberDef));

	md -> pyname = pnd;
	md -> memberflags = 0;
	md -> module = mod;
	md -> next = *flist;

	*flist = md;

	return md;
}


/*
 * Search a set of flags for a particular one and check its type.
 */

static optFlag *findOptFlag(optFlags *flgs,char *name,flagType ft)
{
	int f;

	for (f = 0; f < flgs -> nrFlags; ++f)
	{
		optFlag *of = &flgs -> flags[f];

		if (strcmp(of -> fname,name) == 0)
		{
			/*
			 * An optional name can look like a boolean or a name.
			 */

			if (ft == opt_name_flag)
			{
				if (of -> ftype == bool_flag)
				{
					of -> ftype = opt_name_flag;
					of -> fvalue.sval = NULL;
				}
				else if (of -> ftype == name_flag)
					of -> ftype = opt_name_flag;
			}

			if (ft != of -> ftype)
				yyerror("Optional flag has a value of the wrong type");

			return of;
		}
	}

	return NULL;
}


/*
 * A name is going to be used as a Python attribute name within a Python scope
 * (ie. a Python dictionary), so check against what we already know is going in
 * the same scope in case there is a clash.
 */

static void checkAttributes(sipSpec *pt,classDef *pyscope,nameDef *attr,
			    attrType at)
{
	enumDef *ed;
	varDef *vd;
	classDef *cd;

	/* Check the enum values. */

	for (ed = pt -> enums; ed != NULL; ed = ed -> next)
	{
		enumValueDef *evd;

		if (ed -> ecd != pyscope)
			continue;

		for (evd = ed -> values; evd != NULL; evd = evd -> next)
			if (evd -> name == attr)
				yyerror("There is already an enum value in scope with the same Python name");
	}

	/* Check the variables. */

	for (vd = pt -> vars; vd != NULL; vd = vd -> next)
	{
		if (vd -> ecd != pyscope)
			continue;

		if (vd -> name != attr)
			continue;

		yyerror("There is already a variable in scope with the same Python name");
	}

	/*
	 * Only check the members if this attribute isn't a member because we
	 * can handle members with the same name in the same scope.
	 */

	if (at != func_attr)
	{
		memberDef *md, *membs;

		membs = (pyscope != NULL ? pyscope -> members : pt -> othfuncs);

		for (md = membs; md != NULL; md = md -> next)
		{
			overDef *od, *overs;

			if (md -> pyname != attr)
				continue;

			/* Check for version conflict with all overloads. */

			overs = (pyscope != NULL ? pyscope -> overs : pt -> overs);

			for (od = overs; od != NULL; od = od -> next)
			{
				if (od -> common != md)
					continue;

				yyerror("There is already a function in scope with the same Python name");
			}
		}
	}

	/* Check the classes. */

	for (cd = pt -> classes; cd != NULL; cd = cd -> next)
	{
		if (cd -> ecd != pyscope)
			continue;

		if (cd -> iff -> name != attr)
			continue;

		yyerror("There is already a class or namespace in scope with the same Python name");
	}
}


/*
 * Append a code block to a list of them.  Append is needed to give the
 * specifier easy control over the order of the documentation.
 */

static void appendCodeBlock(codeBlock **headp,codeBlock *new)
{
	while (*headp != NULL)
		headp = &(*headp) -> next;

	*headp = new;
}


/*
 * Handle the end of a fully parsed a file.
 */

static void handleEOF()
{
	/*
	 * Check that the number of nested if's is the same as when we started
	 * the file.
	 */

	if (currentVersScope > newContext.ifdepth)
		fatal("Too many %%If statements in %s\n",previousFile);

	if (currentVersScope < newContext.ifdepth)
		fatal("Too many %%End statements in %s\n",previousFile);
}


/*
 * Handle the end of a fully parsed a module.
 */

static void handleEOM()
{
	/* Check it has been named. */

	if (currentModule -> name == NULL)
		fatal("No %%Module has been specified for module defined in %s\n",previousFile);

	/* The previous module is now current. */

	currentModule = newContext.prevmod;
}


/*
 * Find an existing feature version qualifier.
 */

static versionQual *findFeature(sipSpec *pt,char *name)
{
	versionQual *vq;

	if ((vq = findQualifier(pt,name)) == NULL || vq -> vtype != feature_version)
		yyerror("No such feature");

	return vq;
}
/*
 * Find an existing platform or feature version qualifier.
 */

static versionQual *findPlatOrFeature(sipSpec *pt,char *name)
{
	versionQual *vq;

	if ((vq = findQualifier(pt,name)) == NULL || vq -> vtype == time_version)
		yyerror("No such platform or feature");

	return vq;
}


/*
 * Find an existing version qualifier.
 */

static versionQual *findQualifier(sipSpec *pt,char *name)
{
	moduleDef *mod;

	for (mod = pt -> moduleList; mod != NULL; mod = mod -> next)
	{
		versionQual *vq;

		for (vq = mod -> versions; vq != NULL; vq = vq -> next)
			if (strcmp(vq -> name,name) == 0)
				return vq;
	}

	return NULL;
}


/*
 * Return a copy of a scoped name.
 */

scopedNameDef *copyScopedName(scopedNameDef *snd)
{
	scopedNameDef *head;

	head = NULL;

	while (snd != NULL)
	{
		appendScopedName(&head,text2scopePart(snd -> name));
		snd = snd -> next;
	}

	return head;
}


/*
 * Append a name to a list of scopes.
 */

void appendScopedName(scopedNameDef **headp,scopedNameDef *newsnd)
{
	while (*headp != NULL)
		headp = &(*headp) -> next;

	*headp = newsnd;
}


/*
 * Free a scoped name - but not the text itself.
 */

void freeScopedName(scopedNameDef *snd)
{
	while (snd != NULL)
	{
		scopedNameDef *next = snd -> next;

		free(snd);

		snd = next;
	}
}


/*
 * Convert a text string to a scope part structure.
 */

scopedNameDef *text2scopePart(char *text)
{
	scopedNameDef *snd;

	snd = sipMalloc(sizeof (scopedNameDef));

	snd -> name = text;
	snd -> next = NULL;

	return snd;
}


/*
 * Convert a text string to a fully scoped name.
 */

static scopedNameDef *text2scopedName(char *text)
{
	classDef *cd = currentScope();
	scopedNameDef *snd;

	snd = (cd != NULL ? copyScopedName(cd -> iff -> fqname) : NULL);

	appendScopedName(&snd,text2scopePart(text));

	return snd;
}


/*
 * Return a pointer to the tail part of a scoped name.
 */

char *scopedNameTail(scopedNameDef *snd)
{
	if (snd == NULL)
		return NULL;

	while (snd -> next != NULL)
		snd = snd -> next;

	return snd -> name;
}


/*
 * Push the given scope onto the scope stack.
 */

static void pushScope(classDef *scope)
{
	if (currentScopeIdx >= MAX_NESTED_SCOPE)
		fatal("Internal error: increase the value of MAX_NESTED_SCOPE\n");

	scopeStack[currentScopeIdx] = scope;
	sectFlagsStack[currentScopeIdx] = sectionFlags;

	++currentScopeIdx;
}


/*
 * Pop the scope stack.
 */

static void popScope(void)
{
	if (currentScopeIdx > 0)
		sectionFlags = sectFlagsStack[currentScopeIdx--];
}


/*
 * Set currentVersion according to the requested time and platform and the
 * current features.
 */

static void setCurrentVersion()
{
	int i;

	currentFeature.val = NULL;
	currentVersion = NULL;

	for (i = 0; i < currentVersScope; ++i)
	{
		int passedplat, failedplat;
		versionRange *vr = &versScope[i];
		versionOrList *vol, *newol;

		if (vr -> lower != NULL || vr -> upper != NULL)
		{
			versionQual *known, *vq;

			if ((known = vr -> lower) == NULL)
				known = vr -> upper;

			/* Go through the range in the relevant timeline. */

			for (vq = known -> module -> versions; vq != NULL; vq = vq -> next)
			{
				if (vq -> vtype != time_version)
					continue;

				if (vr -> lower != NULL && vq -> order < vr -> lower -> order)
					continue;

				if (vr -> upper != NULL && vq -> order >= vr -> upper -> order)
					continue;

				if (isNeeded(vq))
					break;
			}

			/* See if any in the range are needed. */

			if (vq == NULL)
				return;
		}

		passedplat = failedplat = FALSE;
		newol = NULL;

		for (vol = vr -> ored; vol != NULL; vol = vol -> next)
			if (vol -> vq -> vtype == platform_version)
				if (isNeeded(vol -> vq))
					passedplat = TRUE;
				else
					failedplat = TRUE;
			else
				orVersionQual(&newol,vol -> vq);

		/*
		 * We don't want this if the expression had a platform we
		 * didn't want but didn't have a platform we did want.
		 */

		if (failedplat && !passedplat)
			return;

		if (newol)
			andVersionOrList(&currentFeature.val,newol);
	}

	/* If we have got this far then the time and platform are Ok. */

	currentVersion = &currentFeature;
}


/*
 * Add a version qualifier to an or'ed list of them.
 */

static void orVersionQual(versionOrList **headp,versionQual *vq)
{
	versionOrList *vol;

	/* See if the qualifier is already included. */

	for (vol = *headp; vol != NULL; vol = vol -> next)
		if (vol -> vq == vq)
			return;

	/* Create the new part of the expression. */

	vol = sipMalloc(sizeof (versionOrList));

	vol -> vq = vq;
	vol -> next = *headp;

	*headp = vol;
}


/*
 * Logically and an or list.
 */

static void andVersionOrList(versionAndList **headp,versionOrList *vol)
{
	versionAndList *val;

	val = sipMalloc(sizeof (versionAndList));

	val -> vol = vol;
	val -> next = NULL;

	/* Find the end of the existing and list. */

	while (*headp)
		headp = &(*headp) -> next;

	*headp = val;
}


/*
 * Return TRUE if the given version qualifier is needed.
 */

static int isNeeded(versionQual *vq)
{
	stringList *sl;

	for (sl = neededVersions; sl != NULL; sl = sl -> next)
		if (strcmp(vq -> name,sl -> s) == 0)
			return TRUE;

	return FALSE;
}


/*
 * Return the current scope.  currentScope() is only valid if currentVersion is
 * non-NULL.
 */

static classDef *currentScope(void)
{
	return (currentScopeIdx > 0 ? scopeStack[currentScopeIdx - 1] : NULL);
}


/*
 * Create a new version qualifier.
 */

static void newVersion(sipSpec *pt,moduleDef *mod,int order,char *name,
		       char *cpp,versionType vt)
{
	versionQual *vq;

	/* Check it doesn't already exist. */

	if (findQualifier(pt,name) != NULL)
		yyerror("Version is already defined");

	vq = sipMalloc(sizeof (versionQual));

	vq -> name = name;
	vq -> vtype = vt;
	vq -> module = mod;
	vq -> order = order;
	vq -> vqcpp = cpp;
	vq -> next = mod -> versions;

	mod -> versions = vq;
}


/*
 * Create a new imported module.
 */

static void newImport(sipSpec *pt,char *name)
{
	/* Ignore if the file has already been imported. */

	moduleDef *mod;

	for (mod = pt -> moduleList; mod != NULL; mod = mod -> next)
		if (strcmp(mod -> file,name) == 0)
			return;

	newModule(NULL,name);
}


/*
 * If the use of an interface is by the main module, add it to the used list.
 */

static void usedInMainModule(sipSpec *pt,ifaceFileDef *iff)
{
	if (isMainModule(iff -> module))
		addToUsedList(&pt -> used,iff);
}
